#!/usr/bin/env python

### BEGIN INIT INFO
# Provides:          tinker-access-client
# Required-Start:    $remote_fs $syslog $network
# Required-Stop:     $remote_fs $syslog
# Default-Start:     2 3 4 5
# Default-Stop:      0 1 6
# Short-Description: tinker-access-client
# Description:       The tinker-access-client service is responsible for coordinating communication
# between RPi's modules (i.e. RFID reader, LCD, power relays etc..) and the remote tinker_access_server.
### END INIT INFO

import sys
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientDaemon import ClientDaemon
from CommandHandler import CommandHandler
from ClientOptionParser import Command


def run():
    logger = ClientLogger.setup()
    logger.debug('Attempting to handle a %s command...', PackageInfo.pip_package_name)
    try:
        with CommandHandler() as handler:
            handler.on(Command.STOP, ClientDaemon.stop)
            handler.on(Command.START, ClientDaemon.start)
            handler.on(Command.STATUS, ClientDaemon.status)
            #handler.on(Command.UPDATE, ClientDaemon.up)
            #handler.on(Command.RESTART, ClientDaemon.restart)
            #handler.on(Command.RELOAD, ClientDaemon.start)
            #handler.on(Command.REMOVE, ClientDaemon.start)
            #handler.on(Command.UNINSTALL, ClientDaemon.start)
            #handler.on(Command.RECONFIGURE, ClientDaemon.start)
            #handler.on(Command.FORCE_RELOAD, ClientDaemon.start)
            handler.wait()

    except (KeyboardInterrupt, SystemExit):
        logger.debug('%s exit.', PackageInfo.pip_package_name)

    except Exception as e:
        logger.debug('%s failed to handle the command.', PackageInfo.pip_package_name)
        logger.exception(e)
        sys.stdout.write(str(e))
        sys.stdout.flush()
        sys.exit(1)


    #TODO: logging.shutDown() might be needed in a finally block

if __name__ == '__main__':
    run()



import os
import sys
import time


# TODO: convert to use LoggedPopen utility...
from subprocess import call, CalledProcessError, check_output

from Client import Client
from Command import Command
from ClientSocket import ClientSocket
from daemonize import Daemonize
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from ClientOption import ClientOption
from ClientOptionParser import ClientOptionParser


# noinspection PyClassHasNoInit


class ClientDaemon:

    @staticmethod
    def start(**kwargs):
        opts = kwargs['opts'] if kwargs['opts'] else {}

        logger = ClientLogger.setup()
        if not ClientDaemon.__status():
            logger.debug('Attempting to start %s...', PackageInfo.pip_package_name)
            try:
                client = Client()
                pid_file = opts.get(ClientOption.PID_FILE)
                foreground = opts.get(ClientOption.DEBUG)
                daemon = Daemonize(
                    app=PackageInfo.pip_package_name,
                    pid=pid_file,
                    action=client.run,
                    foreground=foreground,
                    verbose=True,
                    logger=logger,
                    auto_close_fds=False
                )
                daemon.start()
            except Exception as e:
                logger.debug('%s start failed.', PackageInfo.pip_package_name)
                logger.exception(e)
                raise e
        else:
            sys.stdout.write('{0} is already running...\n'.format(PackageInfo.pip_package_name))
            sys.stdout.flush()
            sys.exit(1)

    @staticmethod
    def stop(**kwargs):
        opts = kwargs['opts'] if kwargs['opts'] else {}

        # Attempt to gracefully shutdown...
        logger = ClientLogger.setup()
        logger.debug('Attempting to stop %s...', PackageInfo.pip_package_name)
        if ClientDaemon.__status():
            ClientDaemon.__stop()

        # if any processes still exists at this point... we will become more persuasive...
        try:
            for pid in check_output(['pgrep', '-f', '{0} start'.format(
                    PackageInfo.pip_package_name)]).splitlines():
                try:
                    logger.debug('Attempting to kill pid: %s', pid)
                    call(['kill', '-9', pid])
                except Exception as e:
                    logger.exception(e)
        except CalledProcessError:
            pass

        # if the pid file still exist at this point... nuke it from orbit!
        pid_file = opts.get(ClientOption.PID_FILE)
        if os.path.isfile(pid_file):
            try:
                logger.debug('Attempting to delete pid_file %s', pid_file)
                os.remove(pid_file)
            except Exception as e:
                logger.exception(e)
                raise e

    @staticmethod
    def restart():
        # TODO: restart should wait until the client is idle.. not in use...

        logger = ClientLogger.setup()
        logger.debug('Attempting to restart %s...', PackageInfo.pip_package_name)
        try:
            ClientDaemon.stop()
            ClientDaemon.start()

        except Exception as e:
            logger.debug('%s restart failed.', PackageInfo.pip_package_name)
            logger.exception(e)
            raise e

    # noinspection PyUnusedLocal
    @staticmethod
    def status(**kwargs):
        logger = ClientLogger.setup()
        logger.debug('Attempting to check %s status...', PackageInfo.pip_package_name)
        status = ClientDaemon.__status()
        if status:
            sys.stdout.write('Status: {0}\n'.format(status))
            sys.stdout.flush()
            sys.exit(0)

        sys.stdout.write('the {0} is unavailable...\n'
                         'It may need to be restarted, try: \'sudo {0} restart\'\n'
                         .format(PackageInfo.pip_package_name))
        sys.stdout.flush()
        sys.exit(1)

    @staticmethod
    def __status():
        # noinspection PyBroadException
        try:
            with ClientSocket() as socket:
                return socket.send(Command.STATUS)

        except Exception:
            pass

        return None

    @staticmethod
    def __stop():
        opts = ClientOptionParser().parse_args()[0]
        logout_coast_time = opts.get(ClientOption.LOGOUT_COAST_TIME)

        # noinspection PyBroadException
        try:
            with ClientSocket(logout_coast_time + 5) as socket:
                socket.send(Command.STOP)

                # TODO: look for ways to exit early here...
                #time.sleep(logout_coast_time)

        except Exception:
            pass

        return None





from CommandHandler import CommandHandler


class RemoteCommandHandler(CommandHandler):
    if __name__ == '__main__':
        def __init__(self):
            super(CommandHandler, self).__init__(False)
            #TODO: init remote listening socket.
            #: move code from client run_listener

    #This will be moved to some other class I am sure...
    def run_listener(self):

        pass

        # self.__logger.debug('Attempting to establish %s listener...', PackageInfo.pip_package_name)
        #
        # server = None
        #
        # try:
        #     server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        #
        #     #TODO: this line causes some major proble, no exception is thrown
        #     #listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        #
        #     server.bind(('', 8089))
        #     server.settimeout(None)
        #     server.listen(5)
        #
        #     now = time.time()
        #     while not self.__should_exit:
        #         self.__logger.debug('listener started...')
        #         client = None
        #
        #         try:
        #             (client, addr) = server.accept()
        #             command = Command(client.recv(1024))
        #             if command:
        #                 self.__logger.debug('%s listener received \'%s\' command',
        #                                     PackageInfo.pip_package_name, command)
        #
        #                 #TODO: allow for graceful shutdown..
        #                 if command is Command.STOP:
        #                     self.__stop()
        #                     break
        #                 else:
        #                     # send a standard response, possibly JSON...
        #                     client.send('State: {0}'.format(self.state))
        #
        #             client.shutdown(socket.SHUT_RDWR)
        #
        #         except socket_error as e:
        #             if e.errno != errno.ENOTCONN:
        #                 self.__logger.exception(e)
        #                 raise e
        #
        #         except Exception as e:
        #             self.__logger.exception(e)
        #             raise e
        #
        #         finally:
        #             if client:
        #                 client.close()
        #
        #         if not self.__should_exit:
        #             self.__logger.debug('will stop listening...')
        #
        # except Exception as e:
        #     self.__logger.debug('Unable to establish the %s listener.', PackageInfo.pip_package_name)
        #     self.__logger.exception(e)
        #     thread.interrupt_main()
        #
        # finally:
        #     if server:
        #         server.shutdown(socket.SHUT_RDWR)
        #         server.close()



import time
import errno
import socket
import thread
from socket import error as socket_error

from transitions import Machine
# TODO: Enable LockedHierarchicalGraphMachine https://github.com/tyarkoni/transitions#-extensions
# from transitions.extensions import LockedHierarchicalGraphMachine as Machine

from Command import Command
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
from DeviceApi import DeviceApi, Channel
from RemoteCommandHandler import RemoteCommandHandler
from ClientOptionParser import ClientOptionParser, ClientOption


class State(object):
    INITIALIZED = 'INITIALIZED'
    IDLE = 'IDLE'


class Client(Machine):

    def __init__(self):
        self.__logger = ClientLogger.setup()
        self.__opts = ClientOptionParser().parse_args()[0]
        self.__should_exit = False

        states = []
        for key, _ in vars(State).items():
            if not key.startswith('__'):
                states.append(key)

        Machine.__init__(self, states=states, initial=State.INITIALIZED)

    def __handle_stop_command(self, **kwargs):
        self.__stop()

    def __stop(self):

        #TODO: wait for client status(self.state) to be done before we exit...

        self.__should_exit = True

    def __handle_status_command(self, **kwargs):
        # TODO: return self.state...
        pass

    def __run(self):
        opts = self.__opts

        with RemoteCommandHandler() as handler, DeviceApi(opts) as device:
            handler.on(Command.STOP, self.__handle_stop_command)
            handler.on(Command.STATUS, self.__handle_status_command)

            def trigger_login():
                self.__logger.debug('trigger login..')

            def trigger_logout():
                if not device.read(Channel.PIN, opts.get(ClientOption.PIN_LOGOUT)):
                    self.__logger.debug('trigger logout..')
                else:
                    thread.Timer(2000, trigger_training_mode).start()

            def trigger_training_mode():
                # If the pin is still HIGH, then go into training mode a.k.a mario mode
                if device.read(Channel.PIN, opts.get(ClientOption.PIN_LOGOUT)):
                    self.__logger.debug('trigger training..')

            device.on(
                Channel.SERIAL,
                direction=device.GPIO.IN,
                call_back=trigger_login
            )

            device.on(
                Channel.PIN,
                pin=opts.get(ClientOption.PIN_LOGOUT),
                direction=device.GPIO.RISING,
                call_back=trigger_logout
            )

            # TODO: are there other things we can sense?
            # i.e. can we do automatic logout request if the machine is shutdown manually?

            while not self.__should_exit:
                self.__logger.debug('%s is waiting...', PackageInfo.pip_package_name)
                device.wait()

    def run(self):
        while not self.__should_exit:
            self.__logger.debug('Attempting to start the %s...', PackageInfo.pip_package_name)
            try:
                self.__run()
            except Exception as e:
                self.__logger.debug('%s failed.', PackageInfo.pip_package_name)
                self.__logger.exception(e)
            self.__logger.debug('Retrying in 5 seconds...')
            time.sleep(5)





import sys
import time
import json
from Command import Command
from PackageInfo import PackageInfo
from ClientLogger import ClientLogger
# from ClientDaemon import ClientDaemon
from ClientOptionParser import ClientOptionParser


# noinspection PyClassHasNoInit
class CommandHandler(object):
    def __init__(self, execute_initial_command=True):
        self.__logger = ClientLogger.setup()
        (opts, args) = ClientOptionParser().parse_args()
        cmd = Command(args[0].lower() if len(args) >= 1 and len(args[0].lower()) >= 1 else None)
        self.__command = (cmd, opts, args) if cmd and execute_initial_command else None
        self.__handlers = []

    # TODO: add test and implement... __enter__, __exit__
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    # def __del__(self):
    #     pass

    def wait(self):
        while not self.__command:
            time.sleep(0.250)

        if self.__handlers:
            for (command, call_back) in self.__handlers:
                (cmd, opts, args) = self.__command
                if command is cmd:
                    call_back(opts=opts, args=args)
            self.__command = None

    def on_command(self, call_back):
        # TODO: implement... probably will need another thread..
        pass

    def on(self, command, call_back):
        self.__handlers.append((command, call_back,))

    # noinspection PyMethodMayBeStatic
    def run(self):
        logger = ClientLogger.setup()
        (opts, args) = ClientOptionParser().parse_args()
        cmd = args[0].lower() if len(args) >= 1 and len(args[0].lower()) >= 1 else None
        command = Command(cmd)

        if command is not None:

            logger.debug(
                'Attempting to handle %s \'%s\' command with:\n%s',
                PackageInfo.pip_package_name, cmd,
                json.dumps(opts, indent=4, sort_keys=True)
            )

            try:
                # #TODO: these should be moved into a command_handler routine inside the ClientDeamom class
                # #since the command handler can be generic now using the on_command call back
                #
                # if command is Command.START:
                #     ClientDaemon.start()
                #
                # elif command is Command.STOP:
                #     ClientDaemon.stop()
                #
                # elif command in [Command.RESTART, Command.RELOAD, Command.FORCE_RELOAD]:
                #     ClientDaemon.restart()
                #
                # elif command is Command.STATUS:
                #     ClientDaemon.status()
                #
                # else:
                #     raise NotImplementedError()

                self.__handle_command(command)

            except (KeyboardInterrupt, SystemExit):
                #  TODO: might not be needed here?
                pass

            except Exception as e:
                logger.debug('%s \'%s\' command failed.', PackageInfo.pip_package_name, cmd)
                logger.exception(e)
                sys.exit(1)



        else:
            raise NotImplementedError

            # except (KeyboardInterrupt, SystemExit):
            #     #  TODO: might not be needed here?
            # pass